# Copyright (C) 2015-2017 Jan Blechta
#
# This file is part of FENaPack.
#
# FENaPack is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FENaPack is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License

"""Compiled extensions for fieldsplit modules"""

from dolfin import compile_cpp_code
import petsc4py

import os

__all__ = ['dofmap_dofs_is', 'SubfieldBC']


dofmap_dofs_is_cpp_code = """
#include <pybind11/pybind11.h>
#include <pybind11/eigen.h>
namespace py = pybind11;

#include <vector>
#include <petscis.h>
#include <dolfin/fem/GenericDofMap.h>
#include <dolfin/la/PETScObject.h>
#include <dolfin/log/log.h>
#include "petsc_casters.h"

IS dofmap_dofs_is(const dolfin::GenericDofMap& dofmap)
{
  PetscErrorCode ierr;
  const std::vector<dolfin::la_index> dofs = dofmap.dofs();
  IS is;
  dolfin_assert(dofmap.index_map());
  ierr = ISCreateGeneral(dofmap.index_map()->mpi_comm(), dofs.size(),
                         dofs.data(), PETSC_COPY_VALUES, &is);
  if (ierr != 0)
    dolfin::PETScObject::petsc_error(ierr, "field_split.py", "ISCreateGeneral");
  return is;
}

PYBIND11_MODULE(SIGNATURE, m)
{
  m.def("dofmap_dofs_is", &dofmap_dofs_is);
}

namespace pybind11
{
  namespace detail
  {
    PETSC_CASTER_MACRO(IS, is);
  }
}
"""

# Load and wrap compiled function dofmap_dofs_is
path = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
module_dofs = compile_cpp_code(dofmap_dofs_is_cpp_code,
                               include_dirs=[path, petsc4py.get_include()])
def dofmap_dofs_is(dofmap):
    """Converts DofMap::dofs() to IS.

    This function is intended to circumvent NumPy which would be
    involved in code like::

        iset = PETSc.IS().createGeneral(dofmap.dofs(),
                                        comm=dofmap.index_map().mpi_comm())
    """
    iset = module_dofs.dofmap_dofs_is(dofmap)
    iset.decRef()
    assert iset.getRefCount() == 1
    return iset
dofmap_dofs_is.__doc__ += module_dofs.dofmap_dofs_is.__doc__


# Load compiled class SubfieldBC
path = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
code = open(os.path.join(path, "SubfieldBC.h")).read()
module_bc = compile_cpp_code(code, include_dirs=[path, petsc4py.get_include()])
SubfieldBC = module_bc.SubfieldBC



# Copyright (C) 2015-2016 Martin Rehor
#
# This file is part of FENaPack.
#
# FENaPack is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FENaPack is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FENaPack.  If not, see <http://www.gnu.org/licenses/>.

from dolfin import Constant, CompiledExpression, FiniteElement, compile_cpp_code

__all__ = ['StabilizationParameterSD']


_streamline_diffusion_cpp = """
#include <pybind11/pybind11.h>
#include <dolfin/function/Expression.h>
#include <dolfin/function/FunctionSpace.h>

using namespace dolfin;


class StabilizationParameterSD : public Expression
{
public:

  std::shared_ptr<GenericFunction> viscosity, density, wind;

  StabilizationParameterSD() : Expression() { }

  void eval(Array<double>& values, const Array<double>& x,
            const ufc::cell& c) const
  {
    // Get dolfin cell and its diameter
    // FIXME: Avoid dynamic allocation
    dolfin_assert(wind->function_space());
    const std::shared_ptr<const Mesh> mesh = wind->function_space()->mesh();
    const Cell cell(*mesh, c.index);
    double h = cell.h();

    // Evaluate viscosity at given coordinates
    // FIXME: Avoid dynamic allocation
    Array<double> nu(viscosity->value_size());
    Array<double> rho(density->value_size());
    viscosity->eval(nu, x, c);
    density->eval(rho, x, c);

    // Compute l2 norm of wind
    double wind_norm = 0.0;
    // FIXME: Avoid dynamic allocation
    Array<double> w(wind->value_size());
    wind->eval(w, x, c);
    for (uint i = 0; i < w.size(); ++i)
      wind_norm += w[i]*w[i];
    wind_norm = sqrt(wind_norm);

    // Compute Peclet number and evaluate stabilization parameter
    double PE = 0.5*wind_norm*h*rho[0]/nu[0];
    values[0] = (PE > 1.0) ? 0.5*h*(1.0 - 1.0/PE)/wind_norm : 0.0;
  }
};

PYBIND11_MODULE(SIGNATURE, m)
{
  pybind11::class_<StabilizationParameterSD,
             std::shared_ptr<StabilizationParameterSD>,
             Expression>
    (m, "StabilizationParameterSD")
    .def(pybind11::init<>())
    .def_readwrite("viscosity", &StabilizationParameterSD::viscosity)
    .def_readwrite("density", &StabilizationParameterSD::density)
    .def_readwrite("wind", &StabilizationParameterSD::wind);
}
"""

_expr = compile_cpp_code(_streamline_diffusion_cpp).StabilizationParameterSD


def StabilizationParameterSD(wind, viscosity, density=None):
    """Returns a subclass of :py:class:`dolfin.Expression` representing
    streamline diffusion stabilization parameter.

    This kind of stabilization is convenient when a multigrid method is used
    for the convection term in the Navier-Stokes equation. The idea of the
    stabilization involves adding an additional term of the form::

      delta_sd*inner(dot(grad(u), w), dot(grad(v), w))*dx

    into the Navier-Stokes equation. Here ``u`` is a trial function, ``v`` is a
    test function and ``w`` defines so-called "wind" which is a known vector
    function.  Regularization parameter ``delta_sd`` is determined by the local
    mesh Peclet number (PE), see the implementation below.

    *Arguments*
        wind (:py:class:`dolfin.GenericFunction`)
            A vector field determining convective velocity.
        viscosity (:py:class:`dolfin.GenericFunction`)
            A scalar field determining dynamic viscosity.
        density (:py:class:`dolfin.GenericFunction`)
            A scalar field determining density (optional).
    """
    if density is None:
        density = Constant(1.0)
    mesh = wind.function_space().mesh()
    element = FiniteElement("DG", mesh.ufl_cell(), 0)
    delta_sd = CompiledExpression(_expr(), element=element, domain=mesh)
    delta_sd.wind = wind._cpp_object
    delta_sd.viscosity = viscosity._cpp_object
    delta_sd.density = density._cpp_object
    return delta_sd



"""Transient flow over a backward-facing step. Incompressible Navier-Stokes
equations are solved using Newton/Picard iterative method. Linear solver is
based on field split PCDR preconditioning."""

# Copyright (C) 2015-2018 Martin Rehor, Jan Blechta
#
# This file is part of FENaPack.
#
# FENaPack is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FENaPack is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FENaPack.  If not, see <http://www.gnu.org/licenses/>.

# Begin demo
from dolfin import *
from matplotlib import pyplot

from fenapack import PCDKrylovSolver
from fenapack import PCDAssembler
from fenapack import PCDNewtonSolver, PCDNonlinearProblem
from fenapack import StabilizationParameterSD

import argparse, sys, os

# Parse input arguments
parser = argparse.ArgumentParser(description=__doc__, formatter_class=
                                 argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument("-l", type=int, dest="level", default=4,
                    help="level of mesh refinement")
parser.add_argument("--nu", type=float, dest="viscosity", default=0.02,
                    help="kinematic viscosity")
parser.add_argument("--pcd", type=str, dest="pcd_variant", default="BRM1",
                    choices=["BRM1", "BRM2"], help="PCD variant")
parser.add_argument("--nls", type=str, dest="nls", default="picard",
                    choices=["picard", "newton"], help="nonlinear solver")
parser.add_argument("--ls", type=str, dest="ls", default="direct",
                    choices=["direct", "iterative"], help="linear solvers")
parser.add_argument("--dm", action='store_true', dest="mumps_debug",
                    help="debug MUMPS")
parser.add_argument("--dt", type=float, dest="dt", default=0.2,
                    help="time step")
parser.add_argument("--t_end", type=float, dest="t_end", default=5.0,
                    help="termination time")
args = parser.parse_args(sys.argv[1:])

# Load mesh from file and refine uniformly
mesh = Mesh(os.path.join(os.path.pardir, "data", "mesh_lshape.xml"))
for i in range(args.level):
    mesh = refine(mesh)

# Define and mark boundaries
class Gamma0(SubDomain):
    def inside(self, x, on_boundary):
        return on_boundary
class Gamma1(SubDomain):
    def inside(self, x, on_boundary):
        return on_boundary and near(x[0], -1.0)
class Gamma2(SubDomain):
    def inside(self, x, on_boundary):
        return on_boundary and near(x[0], 5.0)
boundary_markers = MeshFunction("size_t", mesh, mesh.topology().dim()-1)
boundary_markers.set_all(3)        # interior facets
Gamma0().mark(boundary_markers, 0) # no-slip facets
Gamma1().mark(boundary_markers, 1) # inlet facets
Gamma2().mark(boundary_markers, 2) # outlet facets

# Build Taylor-Hood function space
P2 = VectorElement("Lagrange", mesh.ufl_cell(), 2)
P1 = FiniteElement("Lagrange", mesh.ufl_cell(), 1)
W = FunctionSpace(mesh, P2*P1)

# No-slip BC
bc0 = DirichletBC(W.sub(0), (0.0, 0.0), boundary_markers, 0)

# Parabolic inflow BC
inflow = Expression(("(1.0 - exp(-5.0*t))*4.0*x[1]*(1.0 - x[1])", "0.0"),
                     t=0.0, degree=2)
bc1 = DirichletBC(W.sub(0), inflow, boundary_markers, 1)

# Artificial BC for PCD preconditioner
if args.pcd_variant == "BRM1":
    bc_pcd = DirichletBC(W.sub(1), 0.0, boundary_markers, 1)
elif args.pcd_variant == "BRM2":
    bc_pcd = DirichletBC(W.sub(1), 0.0, boundary_markers, 2)

# Provide some info about the current problem
info("Reynolds number: Re = %g" % (2.0/args.viscosity))
info("Dimension of the function space: %g" % W.dim())

# Arguments and coefficients of the form
u, p = TrialFunctions(W)
v, q = TestFunctions(W)
w = Function(W)
w0 = Function(W)
u_, p_ = split(w)
u0_, p0_ = split(w0)
nu = Constant(args.viscosity)
idt = Constant(1.0/args.dt)

# Nonlinear equation
F = (
      idt*inner(u_ - u0_, v)
    + nu*inner(grad(u_), grad(v))
    + inner(dot(grad(u_), u_), v)
    - p_*div(v)
    - q*div(u_)
)*dx

# Jacobian
if args.nls == "picard":
    J = (
          idt*inner(u, v)
        + nu*inner(grad(u), grad(v))
        + inner(dot(grad(u), u_), v)
        - p*div(v)
        - q*div(u)
    )*dx
elif args.nls == "newton":
    J = derivative(F, w)

# Add stabilization for AMG 00-block
if args.ls == "iterative":
    delta = StabilizationParameterSD(w.sub(0), nu)
    J_pc = J + delta*inner(dot(grad(u), u_), dot(grad(v), u_))*dx
elif args.ls == "direct":
    J_pc = None

# PCD operators
mu = idt*inner(u, v)*dx
mp = Constant(1.0/nu)*p*q*dx
kp = Constant(1.0/nu)*dot(grad(p), u_)*q*dx
ap = inner(grad(p), grad(q))*dx
if args.pcd_variant == "BRM2":
    n = FacetNormal(mesh)
    ds = Measure("ds", subdomain_data=boundary_markers)
    # TODO: What about the reaction term? Does it appear here?
    kp -= Constant(1.0/nu)*dot(u_, n)*p*q*ds(1)
    #kp -= Constant(1.0/nu)*dot(u_, n)*p*q*ds(0)  # TODO: Is this beneficial?

# Collect forms to define nonlinear problem
pcd_assembler = PCDAssembler(J, F, [bc0, bc1],
                             J_pc, ap=ap, kp=kp, mp=mp, mu=mu, bcs_pcd=bc_pcd)
assert pcd_assembler.get_pcd_form("gp").phantom # pressure grad obtained from J
problem = PCDNonlinearProblem(pcd_assembler)

# Set up linear solver (GMRES with right preconditioning using Schur fact)
linear_solver = PCDKrylovSolver(comm=mesh.mpi_comm())
linear_solver.parameters["relative_tolerance"] = 1e-6
#PETScOptions.set("ksp_monitor")
PETScOptions.set("ksp_gmres_restart", 150)

# Set up subsolvers
PETScOptions.set("fieldsplit_p_pc_python_type", "fenapack.PCDRPC_" + args.pcd_variant)
if args.ls == "iterative":
    PETScOptions.set("fieldsplit_u_ksp_type", "richardson")
    PETScOptions.set("fieldsplit_u_ksp_max_it", 1)
    PETScOptions.set("fieldsplit_u_pc_type", "hypre")
    PETScOptions.set("fieldsplit_u_pc_hypre_type", "boomeramg")
    PETScOptions.set("fieldsplit_p_PCD_Rp_ksp_type", "richardson")
    PETScOptions.set("fieldsplit_p_PCD_Rp_ksp_max_it", 1)
    PETScOptions.set("fieldsplit_p_PCD_Rp_pc_type", "hypre")
    PETScOptions.set("fieldsplit_p_PCD_Rp_pc_hypre_type", "boomeramg")
    PETScOptions.set("fieldsplit_p_PCD_Ap_ksp_type", "richardson")
    PETScOptions.set("fieldsplit_p_PCD_Ap_ksp_max_it", 2)
    PETScOptions.set("fieldsplit_p_PCD_Ap_pc_type", "hypre")
    PETScOptions.set("fieldsplit_p_PCD_Ap_pc_hypre_type", "boomeramg")
    PETScOptions.set("fieldsplit_p_PCD_Mp_ksp_type", "chebyshev")
    PETScOptions.set("fieldsplit_p_PCD_Mp_ksp_max_it", 5)
    PETScOptions.set("fieldsplit_p_PCD_Mp_ksp_chebyshev_eigenvalues", "0.5, 2.0")
    #PETScOptions.set("fieldsplit_p_PCD_Mp_ksp_chebyshev_esteig", "1,0,0,1")  # FIXME: What does it do?
    PETScOptions.set("fieldsplit_p_PCD_Mp_pc_type", "jacobi")
elif args.ls == "direct" and args.mumps_debug:
    # Debugging MUMPS
    PETScOptions.set("fieldsplit_u_mat_mumps_icntl_4", 2)
    PETScOptions.set("fieldsplit_p_PCD_Ap_mat_mumps_icntl_4", 2)
    PETScOptions.set("fieldsplit_p_PCD_Mp_mat_mumps_icntl_4", 2)

# Apply options
linear_solver.set_from_options()

# Set up nonlinear solver
solver = PCDNewtonSolver(linear_solver)
solver.parameters["relative_tolerance"] = 1e-5

# Solve problem
t = 0.0
time_iters = 0
krylov_iters = 0
solution_time = 0.0
while t < args.t_end and not near(t, args.t_end, 0.1*args.dt):
    # Move to current time level
    t += args.dt
    time_iters += 1

    # Update boundary conditions
    inflow.t = t

    # Solve the nonlinear problem
    info("t = {:g}, step = {:g}, dt = {:g}".format(t, time_iters, args.dt))
    with Timer("Solve") as t_solve:
        newton_iters, converged = solver.solve(problem, w.vector())
    krylov_iters += solver.krylov_iterations()
    solution_time += t_solve.stop()

    # Update variables at previous time level
    w0.assign(w)

# Report timings
list_timings(TimingClear.clear, [TimingType.wall, TimingType.user])

# Get iteration counts
result = {
    "ndof": W.dim(), "time": solution_time, "steps": time_iters,
    "lin_its": krylov_iters, "lin_its_avg": float(krylov_iters)/time_iters}
tab = "{:^15} | {:^15} | {:^15} | {:^19} | {:^15}\n".format(
    "No. of DOF", "Steps", "Krylov its", "Krylov its (p.t.s.)", "Time (s)")
tab += "{ndof:>9}       | {steps:^15} | {lin_its:^15} | " \
       "{lin_its_avg:^19.1f} | {time:^15.2f}\n".format(**result)
print("\nSummary of iteration counts:")
print(tab)
#with open("table_pcdr_{}.txt".format(args.pcd_variant), "w") as f:
#    f.write(tab)

# Plot solution
u, p = w.split()
size = MPI.size(mesh.mpi_comm())
rank = MPI.rank(mesh.mpi_comm())
pyplot.figure()
pyplot.subplot(2, 1, 1)
plot(u, title="velocity")
pyplot.subplot(2, 1, 2)
plot(p, title="pressure")
pyplot.savefig("figure_v_p_size{}_rank{}.pdf".format(size, rank))
pyplot.figure()
plot(p, title="pressure", mode="warp")
pyplot.savefig("figure_warp_size{}_rank{}.pdf".format(size, rank))
pyplot.show()



# Copyright (C) 2015-2018 Jan Blechta and Martin Rehor
#
# This file is part of FENaPack.
#
# FENaPack is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FENaPack is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FENaPack.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides wrappers for assembling systems of linear algebraic
equations to be solved with the use of PCD preconditioning strategy.
These wrappers naturally provide routines for assembling preconditioning
operators themselves.
"""

from dolfin import SystemAssembler, assemble


class PCDAssembler(object):
    """Base class for creating linear problems to be solved by application
    of the PCD preconditioning strategy. Users are encouraged to use this class
    for interfacing with :py:class:`fenapack.field_split.PCDKrylovSolver`.
    On request it assembles not only the individual PCD operators but also the
    system matrix and the right hand side vector defining the linear problem.
    """

    def __init__(self, a, L, bcs, a_pc=None,
                 mp=None, mu=None, ap=None, fp=None, kp=None, gp=None,
                 bcs_pcd=[]):
        """Collect individual variational forms and boundary conditions
        defining a linear problem (system matrix + RHS vector) on the one side
        and preconditioning operators on the other side.

        *Arguments*
            a (:py:class:`dolfin.Form` or :py:class:`ufl.Form`)
                Bilinear form representing a system matrix.
            L (:py:class:`dolfin.Form` or :py:class:`ufl.Form`)
                Linear form representing a right hand side vector.
            bcs (:py:class:`list` of :py:class:`dolfin.DirichletBC`)
                Boundary conditions applied to ``a``, ``L``, and ``a_pc``.
            a_pc (:py:class:`dolfin.Form` or :py:class:`ufl.Form`)
                Bilinear form representing a matrix optionally passed to
                preconditioner instead of ``a``. In case of PCD, stabilized
                00-block can be passed to 00-KSP solver.
            mp, mu, ap, fp, kp, gp (:py:class:`dolfin.Form` or :py:class:`ufl.Form`)
                Bilinear forms which (some of them) might be used by a
                particular PCD(R) preconditioner. Typically they represent "mass
                matrix" on pressure, "mass matrix" on velocity, minus Laplacian
                operator on pressure, pressure convection-diffusion operator,
                pressure convection operator and pressure gradient respectively.
            bcs_pcd (:py:class:`list` of :py:class:`dolfin.DirichletBC`)
                Artificial boundary conditions used by PCD preconditioner.

        All the arguments should be given on the common mixed function space.

        All the forms are wrapped using :py:class:`PCDForm` so that each of
        them can be endowed with additional set of properties.

        By default, ``mp``, ``mu``, ``ap`` and ``gp`` are assumed to be
        constant if the preconditioner is used repeatedly in some outer
        iterative process (e.g Newton-Raphson method, time-stepping).
        As such, the corresponding operators are assembled only once.
        On the other hand, ``fp`` and ``kp`` are updated in every
        outer iteration.

        Also note that ``gp`` is the only form that is by default in a *phantom
        mode*. It means that the corresponding operator (if needed) is not
        obtained by assembling the form, but it is extracted as the 01-block of
        the system matrix.

        The default setting can be modified by accessing
        a :py:class:`PCDForm` instance via :py:meth:`PCDAssembler.get_pcd_form`
        and changing the properties directly.
        """

        # Assembler for the linear system of algebraic equations
        self.assembler = SystemAssembler(a, L, bcs)

        # Assembler for preconditioner
        if a_pc is not None:
            self.assembler_pc = SystemAssembler(a_pc, L, bcs)
        else:
            self.assembler_pc = None

        # Store bcs
        self._bcs = bcs
        self._bcs_pcd = bcs_pcd

        # Store and initialize forms
        self._forms = {
            "L": PCDForm(L),
            "ap": PCDForm(ap, const=True),
            "mp": PCDForm(mp, const=True),
            "mu": PCDForm(mu, const=True),
            "fp": PCDForm(fp),
            "kp": PCDForm(kp),
            "gp": PCDForm(gp, const=True, phantom=True),
        }


    def get_pcd_form(self, key):
        """Return form wrapped in :py:class:`PCDForm`."""
        form = self._forms.get(key)
        if form is None:
            raise AttributeError("Form '%s' requested by PCD not available" % key)
        assert isinstance(form, PCDForm)
        return form


    def get_dolfin_form(self, key):
        """Return form as :py:class:`dolfin.Form` or :py:class:`ufl.Form`."""
        return self.get_pcd_form(key).dolfin_form()


    def function_space(self):
        return self.get_dolfin_form("L").arguments()[0].function_space()


    def rhs_vector(self, b, x=None):
        """Assemble right hand side vector ``b``.

        The version with ``x`` is suitable for use inside
        a (quasi)-Newton solver.
        """
        if x is not None:
            self.assembler.assemble(b, x)
        else:
            self.assembler.assemble(b)

    def system_matrix(self, A):
        """Assemble system matrix ``A``."""
        self.assembler.assemble(A)


    def pc_matrix(self, P):
        """Assemble preconditioning matrix ``P`` whose relevant blocks can be
        passed to actual parts of the ``KSP`` solver.
        """
        if self.assembler_pc is not None:
            self.assembler_pc.assemble(P)


    def ap(self, Ap):
        assembler = SystemAssembler(self.get_dolfin_form("ap"),
                                    self.get_dolfin_form("L"),
                                    self.pcd_bcs())
        assembler.assemble(Ap)


    def mp(self, Mp):
        assemble(self.get_dolfin_form("mp"), tensor=Mp)


    def mu(self, Mu):
        assemble(self.get_dolfin_form("mu"), tensor=Mu)


    def fp(self, Fp):
        assemble(self.get_dolfin_form("fp"), tensor=Fp)


    def kp(self, Kp):
        assemble(self.get_dolfin_form("kp"), tensor=Kp)


    def gp(self, Bt):
        """Assemble discrete pressure gradient. It is crucial to respect any
        constraints placed on the velocity test space by Dirichlet boundary
        conditions."""
        assemble(self.get_dolfin_form("gp"), tensor=Bt)
        for bc in self._bcs:
            bc.apply(Bt)


    # FIXME: Naming
    def pcd_bcs(self):
        try:
            assert self._bcs_pcd is not None
        except (AttributeError, AssertionError):
            raise AttributeError("BCs requested by PCD not available")
        return self._bcs_pcd


class PCDForm(object):
    """Wrapper for PCD operators represented by :py:class:`dolfin.Form` or
    :py:class:`ufl.Form`. This class allows to record specific properties of
    the form that can be utilized later while setting up the preconditioner.

    For example, we can specify which matrices remain constant during the outer
    iterative algorithm (e.g. Newton-Raphson method, time-stepping) and which
    matrices need to be updated in every outer iteration.
    """
    def __init__(self, form, const=False, phantom=False):
        """The class is initialized by a single form with default properties.

        *Arguments*
            form (:py:class:`dolfin.Form` or :py:class:`ufl.Form`)
                A form to be wrapped.
            const (`bool`)
                Whether the form remains constant in outer iterations.
            phantom (`bool`)
                If `True`, then the corresponding operator will be obtained
                not by assembling the form, but in some different way. (For
                example, pressure gradient may be extracted directly from
                the system matrix.)
        """
        # Store form
        self._form = form

        # Initialize public properties
        assert isinstance(const, bool)
        self.constant = const
        self.phantom = phantom

    def dolfin_form(self):
        return self._form

    def is_constant(self):
        return self.constant

    def is_phantom(self):
        return self.phantom



# Copyright (C) 2017, 2018 Jan Blechta, Martin Rehor
#
# This file is part of FENaPack.
#
# FENaPack is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FENaPack is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License

"""Tools for extraction and management of fieldsplit
submatrices, subvectors, subbcs, subksps intended to
be hidden from user interface"""

from dolfin import PETScMatrix, Timer, DirichletBC
from petsc4py import PETSc

from fenapack._field_split_utils import SubfieldBC
from fenapack.assembling import PCDAssembler

from weakref import WeakKeyDictionary, proxy


class PCDInterface(object):
    """Wrapper of PCDAssembler for interfacing with PCD PC
    fieldsplit implementation. Convection fieldsplit submatrices
    are extracted as shallow or deep submatrices according to
    ``deep_submats`` parameter."""

    def __init__(self, pcd_assembler, A, is_u, is_p, deep_submats=False):
        """Create PCDInterface instance given PCDAssembler instance,
        system matrix and velocity and pressure index sets"""

        # Check input
        assert isinstance(pcd_assembler, PCDAssembler)
        assert isinstance(is_u, PETSc.IS)
        assert isinstance(is_p, PETSc.IS)

        # Store what needed
        self.assembler = pcd_assembler
        self.A = proxy(A)
        self.is_u = is_u
        self.is_p = is_p

        # Choose submatrix implementation
        assert isinstance(deep_submats, bool)
        if deep_submats:
            self.assemble_operator = self._assemble_operator_deep
        else:
            self.assemble_operator = self._assemble_operator_shallow

        # Dictionary for storing work mats
        self.scratch = WeakKeyDictionary()


    def apply_pcd_bcs(self, vec):
        """Apply bcs to intermediate pressure vector of PCD pc"""
        self.apply_bcs(vec, self.assembler.pcd_bcs, self.is_p)


    def setup_ksp_Ap(self, ksp):
        """Setup pressure Laplacian ksp and assemble matrix"""
        self.setup_ksp(ksp, self.assembler.ap, self.is_p, spd=True,
                       const=self.assembler.get_pcd_form("ap").is_constant())


    def setup_ksp_Mp(self, ksp):
        """Setup pressure mass matrix ksp and assemble matrix"""
        self.setup_ksp(ksp, self.assembler.mp, self.is_p, spd=True,
                       const=self.assembler.get_pcd_form("mp").is_constant())


    def setup_mat_Kp(self, mat=None):
        """Setup and assemble pressure convection
        matrix and return it"""
        if mat is None or not self.assembler.get_pcd_form("kp").is_constant():
            return self.assemble_operator(self.assembler.kp, self.is_p, submat=mat)


    def setup_mat_Fp(self, mat=None):
        """Setup and assemble pressure convection-diffusion
        matrix and return it"""
        if mat is None or not self.assembler.get_pcd_form("fp").is_constant():
            return self.assemble_operator(self.assembler.fp, self.is_p, submat=mat)


    def setup_mat_Mu(self, mat=None):
        """Setup and assemble velocity mass matrix
        and return it"""
        # NOTE: deep submats are required for the later use in _build_approx_Ap
        if mat is None or not self.assembler.get_pcd_form("mu").is_constant():
            return self._assemble_operator_deep(self.assembler.mu, self.is_u, submat=mat)


    def setup_mat_Bt(self, mat=None):
        """Setup and assemble discrete pressure gradient
        and return it"""
        # NOTE: deep submats are required for the later use in _build_approx_Ap
        if mat is None or not self.assembler.get_pcd_form("gp").is_constant():
            if self.assembler.get_pcd_form("gp").is_phantom():
                # NOTE: Bt is obtained from the system matrix
                return self._get_deep_submat(self.A, self.is_u, self.is_p, submat=mat)
            else:
                # NOTE: Bt is obtained by assembling a form
                return self._assemble_operator_deep(self.assembler.gp,
                                                    self.is_u, self.is_p, submat=mat)


    def setup_ksp_Rp(self, ksp, Mu, Bt):
        """Setup pressure Laplacian ksp based on velocity mass matrix ``Mu``
        and discrete gradient ``Bt`` and assemble matrix
        """
        mat = ksp.getOperators()[0]
        prefix = ksp.getOptionsPrefix()
        const = self.assembler.get_pcd_form("mu").is_constant() \
                  and self.assembler.get_pcd_form("gp").is_constant()
        if mat.type is None or not mat.isAssembled() or not const:
            # Get approximate Laplacian
            mat = self._build_approx_Ap(Mu, Bt, mat)

            # Use eventual spd flag
            mat.setOption(PETSc.Mat.Option.SPD, True)

            # Set correct options prefix
            mat.setOptionsPrefix(prefix)

            # Use also as preconditioner matrix
            ksp.setOperators(mat, mat)
            assert ksp.getOperators()[0].isAssembled()

            # Setup ksp
            with Timer("FENaPack: {} setup".format(prefix)):
                ksp.setUp()


    def _build_approx_Ap(self, Mu, Bt, mat=None):
        # Fetch work vector and matrix
        diagMu, = self.get_work_vecs_from_square_mat(Mu, 1)
        Ap, = self.get_work_mats(Bt, 1)

        # Get diagonal of the velocity mass matrix
        Mu.getDiagonal(result=diagMu)

        # Make inverse of diag(Mu)
        diagMu.reciprocal() # diag(Mu)^{-1}

        # Make square root of the diagonal and use it for scaling
        diagMu.sqrtabs() # \sqrt{diag(Mu)^{-1}}

        # Process discrete "grad" operator
        Bt.copy(result=Ap)         # Ap = Bt
        Ap.diagonalScale(L=diagMu) # scale rows of Ap, i.e. Ap = diagMu*Bt

        # Return Ap = Ap^T*Ap, which is B diag(Mu)^{-1} B^T,
        if mat is None or not mat.isAssembled():
            return Ap.transposeMatMult(Ap)
        else:
            # NOTE: 'result' can only be used if the multiplied matrices have
            #       the same nonzero pattern as in the previous call
            return Ap.transposeMatMult(Ap, result=mat)


    def get_work_vecs_from_square_mat(self, M, num):
        """Return ``num`` of work vecs initially created from a square
        matrix ``M``."""
        # Verify that we have a square matrix
        m, n = M.getSize()
        assert m == n
        try:
            vecs = self._work_vecs
            assert len(vecs) == num
        except AttributeError:
            self._work_vecs = vecs = tuple(M.getVecLeft() for i in range(num))
        except AssertionError:
            raise ValueError("Changing number of work vecs not allowed")
        return vecs


    def get_work_mats(self, M, num):
        """Return ``num`` of work mats initially created from matrix ``B``."""
        try:
            mats = self._work_mats
            assert len(mats) == num
        except AttributeError:
            self._work_mats = mats = tuple(M.duplicate() for i in range(num))
        except AssertionError:
            raise ValueError("Changing number of work mats not allowed")
        return mats


    def get_work_dolfin_mat(self, key, comm,
                            can_be_destroyed=None, can_be_shared=None):
        """Get working DOLFIN matrix by key. ``can_be_destroyed=True`` tells
        that it is probably favourable to not store the matrix unless it is
        shared as it will not be used ever again, ``None`` means that it can
        be destroyed but it is not probably favourable and ``False`` forbids
        the destruction. ``can_be_shared`` tells if a work matrix can be the
        same with work matrices for other keys."""
        # TODO: Add mechanism for sharing DOLFIN mats
        # NOTE: Maybe we don't really need sharing. If only persistent matrix
        #       is convection then there is nothing to be shared.

        # Check if requested matrix is in scratch
        dolfin_mat = self.scratch.get(key, None)

        # Allocate new matrix otherwise
        if dolfin_mat is None:

            if isinstance(comm, PETSc.Comm):
                comm = comm.tompi4py()

            dolfin_mat = PETScMatrix(comm)

        # Store or pop the matrix as requested
        if can_be_destroyed in [False, None]:
            self.scratch[key] = dolfin_mat
        else:
            assert can_be_destroyed is True
            self.scratch.pop(key, None)

        return dolfin_mat


    def setup_ksp(self, ksp, assemble_func, iset, spd=False, const=False):
        """Assemble into operator of given ksp if not yet assembled"""
        mat = ksp.getOperators()[0]
        prefix = ksp.getOptionsPrefix()
        if mat.type is None or not mat.isAssembled():
            # Assemble matrix
            destruction = True if const else None
            dolfin_mat = self.get_work_dolfin_mat(assemble_func, mat.comm,
                                                  can_be_destroyed=destruction,
                                                  can_be_shared=True)
            assemble_func(dolfin_mat)
            mat = self._get_deep_submat(dolfin_mat.mat(), iset, submat=None)

            # Use eventual spd flag
            mat.setOption(PETSc.Mat.Option.SPD, spd)

            # Set correct options prefix
            mat.setOptionsPrefix(prefix)

            # Use also as preconditioner matrix
            ksp.setOperators(mat, mat)
            assert ksp.getOperators()[0].isAssembled()

            # Set up ksp
            with Timer("FENaPack: {} setup".format(prefix)):
                ksp.setUp()

        elif not const:
            # Assemble matrix and set up ksp
            mat = self._assemble_operator_deep(assemble_func, iset, submat=mat)
            assert mat.getOptionsPrefix() == prefix
            ksp.setOperators(mat, mat)
            with Timer("FENaPack: {} setup".format(prefix)):
                ksp.setUp()


    def _assemble_operator_shallow(self, assemble_func, isrow, iscol=None, submat=None):
        """Assemble operator of given name using shallow submat"""
        # Assemble into persistent DOLFIN matrix everytime
        # TODO: Does not shallow submat take care of parents lifetime? How?
        dolfin_mat = self.get_work_dolfin_mat(assemble_func, isrow.comm,
                                              can_be_destroyed=False,
                                              can_be_shared=False)
        assemble_func(dolfin_mat)

        # FIXME: This logic that it is created once should be visible
        #        in higher level, not in these internals
        # Create shallow submatrix (view into dolfin mat) once
        if submat is None or submat.type is None or not submat.isAssembled():
            submat = self._get_shallow_submat(dolfin_mat.mat(), isrow, iscol, submat=submat)
            assert submat.isAssembled()

        return submat


    def _assemble_operator_deep(self, assemble_func, isrow, iscol=None, submat=None):
        """Assemble operator of given name using deep submat"""
        dolfin_mat = self.get_work_dolfin_mat(assemble_func, isrow.comm,
                                              can_be_destroyed=None,
                                              can_be_shared=True)
        assemble_func(dolfin_mat)
        return self._get_deep_submat(dolfin_mat.mat(), isrow, iscol, submat=submat)


    def apply_bcs(self, vec, bcs_getter, iset):
        """Transform dolfin bcs obtained using ``bcs_getter`` function
        into fieldsplit subBCs and apply them to fieldsplit vector.
        SubBCs are cached."""
        # Fetch subbcs from cache or construct it
        subbcs = getattr(self, "_subbcs", None)
        if subbcs is None:
            bcs = bcs_getter()
            bcs = [bcs] if isinstance(bcs, DirichletBC) else bcs
            subbcs = [SubfieldBC(bc, iset) for bc in bcs]
            self._subbcs = subbcs

        # Apply bcs
        for bc in subbcs:
            bc.apply(vec)


    if PETSc.Sys.getVersion()[0:2] <= (3, 7) and PETSc.Sys.getVersionInfo()['release']:

        @staticmethod
        def _get_deep_submat(mat, isrow, iscol=None, submat=None):
            if iscol is None:
                iscol = isrow
            return mat.getSubMatrix(isrow, iscol, submat=submat)

        @staticmethod
        def _get_shallow_submat(mat, isrow, iscol=None, submat=None):
            if iscol is None:
                iscol = isrow
            if submat is None:
                submat = PETSc.Mat().create(isrow.comm)
            return submat.createSubMatrix(mat, isrow, iscol)


    else:

        @staticmethod
        def _get_deep_submat(mat, isrow, iscol=None, submat=None):
            if iscol is None:
                iscol = isrow
            return mat.createSubMatrix(isrow, iscol, submat=submat)

        @staticmethod
        def _get_shallow_submat(mat, isrow, iscol=None, submat=None):
            if iscol is None:
                iscol = isrow
            if submat is None:
                submat = PETSc.Mat().create(isrow.comm)
            return submat.createSubMatrixVirtual(mat, isrow, iscol)



from __future__ import print_function

from dolfin import *
from mshr import *

import os
import sys

path = os.path.join(os.getcwd(), os.path.dirname(__file__))
path = os.path.realpath(path)

resolution = int(sys.argv[1])

end = 150
ratio = 6
inlet = Rectangle(Point(0, -1), Point(2.5, 1))
main  = Rectangle(Point(2.5, -ratio), Point(end, +ratio))
domain = inlet + main

m = generate_mesh(domain, resolution)

print("num vertices:", m.num_vertices())
print("num cells:", m.num_cells())

File(os.path.join(path, "mesh.xml.gz")) << m



# Copyright (C) 2015-2017 Jan Blechta and Martin Rehor
#
# This file is part of FENaPack.
#
# FENaPack is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FENaPack is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FENaPack.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides subclasses of DOLFIN interface for
solving non-linear problems suitable for use with fieldsplit
preconditioned Krylov methods
"""

from dolfin import NewtonSolver, PETScFactory, NonlinearProblem

from fenapack.assembling import PCDAssembler


class PCDNewtonSolver(NewtonSolver):
    """Newton solver suitable for use with
    :py:class:`fenapack.field_split.PCDKrylovSolver`.
    """

    def __init__(self, solver, pcd_pc_class=None):
        """Initialize for a given PCD Krylov solver and preconditioner class.

        *Arguments*
            solver (:py:class:`fenapack.field_split.PCDKrylovSolver`)
               A PCD Krylov solver.
            pcd_pc_class (:py:class:`fenapack.preconditioners.BasePCDPC`)
               Representative of a class implementing PCD preconditioner.
        """

        # Initialize DOLFIN Newton solver
        comm = solver.ksp().comm.tompi4py()
        factory = PETScFactory.instance()
        super(PCDNewtonSolver, self).__init__(comm, solver, factory)

        # Store Python reference for solver setup
        self._solver = solver
        self._pcd_pc_class = pcd_pc_class


    def solve(self, problem, x):
        # Store Python reference for solver setup
        self._problem = problem

        # Solve the problem, drop the reference, and return
        r = super(PCDNewtonSolver, self).solve(problem, x)
        del self._problem
        return r


    def solver_setup(self, A, P, nonlinear_problem, iteration):
        # Only do the setup once
        # FIXME: Is this good?
        if iteration > 0 or getattr(self, "_initialized", False):
            return
        self._initialized = True

        # C++ references passed in do not have Python context
        linear_solver = self._solver
        nonlinear_problem = self._problem

        # Set operators and initialize PCD
        P = A if P.empty() else P
        linear_solver.set_operators(A, P)
        linear_solver.init_pcd(
            nonlinear_problem.pcd_assembler, self._pcd_pc_class)


    def linear_solver(self):
        return self._solver


class PCDNonlinearProblem(NonlinearProblem):
    """Class for interfacing with :py:class:`PCDNewtonSolver`."""

    def __init__(self, pcd_assembler):
        """Return subclass of :py:class:`dolfin.NonlinearProblem`
        suitable for :py:class:`NewtonSolver` based on
        :py:class:`fenapack.field_split.PCDKrylovSolver` and
        PCD preconditioners from :py:class:`fenapack.preconditioners`.

        *Arguments*
            pcd_assembler (:py:class:`fenapack.assembling.PCDAssembler`)
               A class which takes care of assembling PCD operators on demand.
        """

        assert isinstance(pcd_assembler, PCDAssembler)
        super(PCDNonlinearProblem, self).__init__()
        self.pcd_assembler = pcd_assembler

    def F(self, b, x):
        self.pcd_assembler.rhs_vector(b, x)


    def J(self, A, x):
        self.pcd_assembler.system_matrix(A)


    def J_pc(self, P, x):
        self.pcd_assembler.pc_matrix(P)



# Copyright (C) 2014-2017 Jan Blechta and Martin Rehor
#
# This file is part of FENaPack.
#
# FENaPack is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FENaPack is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FENaPack.  If not, see <http://www.gnu.org/licenses/>.

from dolfin import timed
from petsc4py import PETSc

from fenapack.utils import get_default_factor_solver_type
from fenapack.utils import pc_set_factor_solver_type


class BasePCDPC(object):
    """Base python context for pressure convection diffusion (PCD)
    preconditioners."""
    def create(self, pc):
        self.ksp_Ap = self.create_default_ksp(pc.comm)
        self.ksp_Mp = self.create_default_ksp(pc.comm)

        options_prefix = pc.getOptionsPrefix()
        self.ksp_Ap.setOptionsPrefix(options_prefix + "PCD_Ap_")
        self.ksp_Mp.setOptionsPrefix(options_prefix + "PCD_Mp_")


    def setFromOptions(self, pc):
        self.ksp_Ap.setFromOptions()
        self.ksp_Mp.setFromOptions()


    @staticmethod
    def create_default_ksp(comm):
        """Return Cholesky factorization KSP"""
        ksp = PETSc.KSP().create(comm)
        ksp.setType(PETSc.KSP.Type.PREONLY)
        ksp.pc.setType(PETSc.PC.Type.CHOLESKY)
        pc_set_factor_solver_type(ksp.pc, get_default_factor_solver_type(comm))
        return ksp


    def get_work_vecs(self, v, num):
        """Return ``num`` of work vecs initially duplicated from v"""
        try:
            vecs = self._work_vecs
            assert len(vecs) == num
        except AttributeError:
            self._work_vecs = vecs = tuple(v.duplicate() for i in range(num))
        except AssertionError:
            raise ValueError("Changing number of work vecs not allowed")
        return vecs


    def init_pcd(self, pcd_interface):
        """Initialize by PCDInterface instance"""
        if hasattr(self, "interface"):
            raise RuntimeError("Reinitialization of PCDPC not allowed")
        self.interface = pcd_interface


    def setUp(self, pc):
        # Prepare mass matrix and Laplacian solvers
        # NOTE: Called function ensures that assembly, submat extraction and
        #       ksp setup is done only once during preconditioner lifetime.
        self.interface.setup_ksp_Mp(self.ksp_Mp)
        self.interface.setup_ksp_Ap(self.ksp_Ap)

        # Prepare convection matrix
        Kp = self.interface.setup_mat_Kp(mat=getattr(self, "mat_Kp", None))
        if Kp is not None: # updated only if not constant
            self.mat_Kp = Kp
            self.mat_Kp.setOptionsPrefix(pc.getOptionsPrefix() + "PCD_Kp_")

        # Fetch bcs apply function
        self.bcs_applier = self.interface.apply_pcd_bcs



class PCDPC_BRM1(BasePCDPC):
    """This class implements a modification of PCD variant similar to one by
    [1]_.

    .. [1] Olshanskii M. A., Vassilevski Y. V., *Pressure Schur complement
           preconditioners for the discrete Oseen problem*.
           SIAM J. Sci. Comput., 29(6), 2686-2704. 2007.
    """

    @timed("FENaPack: PCDPC_BRM1 apply")
    def apply(self, pc, x, y):
        r"""This method implements the action of the inverse of the approximate
        Schur complement :math:`-\hat{S}^{-1}`, that is

        .. math::

            y = -M_p^{-1} (I + K_p A_p^{-1}) x

        where :math:`K_p` is used to denote pressure convection matrix plus
        possibly pressure mass matrix coming from discrete time derivative.
        Note that Laplace solve with :math:`A_p^{-1} x` is performed with usual
        non-symmetric application of subfield BC on matrix :math:`A_p` and RHS
        :math:`x` (but only in that term). It is crucial that identity term
        :math:`I x` is not absorbed into the second, compact term to get

        .. math::

            y = -M_p^{-1} (A_p + K_p) A_p^{-1} x.

        This is crucial to keep a stability with respect to the leading Stokes
        term.

        Good strategy is to use :math:`M_p` and :math:`K_p` both scaled by
        :math:`\nu^{-1}`.
        """
        # Fetch work vector
        z, = self.get_work_vecs(x, 1)

        # Apply PCD
        x.copy(result=z)        # z = x
        self.bcs_applier(z)     # apply bcs to z
        self.ksp_Ap.solve(z, y) # y = A_p^{-1} z
        self.mat_Kp.mult(y, z)  # z = K_p y
        z.axpy(1.0, x)          # z = z + x
        self.ksp_Mp.solve(z, y) # y = M_p^{-1} z
        # FIXME: How is with the sign bussines?
        y.scale(-1.0)           # y = -y



class PCDPC_BRM2(BasePCDPC):
    """This class implements a modification of steady variant of PCD
    described in [2]_.

    .. [2] Elman H. C., Silvester D. J., Wathen A. J., *Finite Elements and Fast
           Iterative Solvers: With Application in Incompressible Fluid Dynamics*.
           Oxford University Press 2005. 2nd edition 2014.
    """

    @timed("FENaPack: PCDPC_BRM2 apply")
    def apply(self, pc, x, y):
        # FIXME: Fix the docstring
        """This method implements the action of the inverse of the approximate
        Schur complement :math:`\hat{S} = -M_p F_p^{-1} A_p`, that is

        .. math::

            y = \hat{S}^{-1} x = - (I + A_p^{-1} K_p) M_p^{-1} x.
        """
        # Fetch work vector
        z0, z1 = self.get_work_vecs(x, 2)

        # Apply PCD
        self.ksp_Mp.solve(x, y)   # y = M_p^{-1} x
        y.copy(result=z0)         # z0 = y
        self.mat_Kp.mult(z0, z1)  # z1 = K_p z0
        self.bcs_applier(z1)      # apply bcs to z1
        self.ksp_Ap.solve(z1, z0) # z0 = A_p^{-1} z1
        y.axpy(1.0, z0)           # y = y + z0
        # FIXME: How is with the sign bussines?
        y.scale(-1.0)             # y = -y



class BasePCDRPC(BasePCDPC):
    """Base python context for pressure convection diffusion reaction (PCDR)
    preconditioners.
    """
    # TODO: Add demo/bench to demonstrate performance of PCDR
    def create(self, pc):
        super(BasePCDRPC, self).create(pc)
        self.ksp_Rp = self.create_default_ksp(pc.comm)

        options_prefix = pc.getOptionsPrefix()
        self.ksp_Rp.setOptionsPrefix(options_prefix + "PCD_Rp_")


    def setFromOptions(self, pc):
        super(BasePCDRPC, self).setFromOptions(pc)
        self.ksp_Rp.setFromOptions()


    def setUp(self, pc):
        # Prepare mass matrix and Laplacian solvers, convection matrix and bcs_applier
        super(BasePCDRPC, self).setUp(pc)

        # Prepare Laplacian solver based on velocity mass matrix
        # and discrete pressure gradient
        Mu = self.interface.setup_mat_Mu(mat=getattr(self, "mat_Mu", None))
        if Mu is not None: # updated only if not constant
            self.mat_Mu = Mu
            self.mat_Mu.setOptionsPrefix(pc.getOptionsPrefix() + "PCD_Mu_")

        Bt = self.interface.setup_mat_Bt(mat=getattr(self, "mat_Bt", None))
        if Bt is not None: # updated only if not constant
            self.mat_Bt = Bt
            self.mat_Bt.setOptionsPrefix(pc.getOptionsPrefix() + "PCD_Bt_")

        self.interface.setup_ksp_Rp(self.ksp_Rp, self.mat_Mu, self.mat_Bt)



class PCDRPC_BRM1(BasePCDRPC):
    """This class implements an extension of :py:class:`PCDPC_BRM1`.
    Here we add a reaction term into the preconditioner, so that is becomes
    PCDR (pressure-convection-diffusion-reaction) preconditioner. This
    particular variant is suitable for time-dependent problems, where the
    reaction term arises from the time derivative in the balance of momentum.
    """

    @timed("FENaPack: PCDRPC_BRM1 apply")
    def apply(self, pc, x, y):
        r"""This method implements the action of the inverse of the approximate
        Schur complement :math:`-\hat{S}^{-1}`, that is

        .. math::

            y = - R_p^{-1} x - M_p^{-1} (I + K_p A_p^{-1}) x

        where :math:`K_p` is used to denote pressure convection matrix, while
        :math:`R_p` originates in the discretized time derivative. Roughly
        speaking, :math:`R_p^{-1} x` corresponds to additional Laplace solve
        in which the minus Laplacian operator is approximated in a clever way
        based on the *velocity mass matrix* and *pressure gradient operator*.

        Based on experimental evidence, we can say that this particular
        variant performs better than :py:class:`PCDPC_BRM1` (with :math:`K_p`
        enriched by the pressure mass matrix from the discrete time
        derivative) applied to time-dependent problems.
        """
        # Fetch work vector
        z, = self.get_work_vecs(x, 1)

        # Apply PCD
        x.copy(result=z)        # z = x
        self.bcs_applier(z)     # apply bcs to z
        self.ksp_Ap.solve(z, y) # y = A_p^{-1} z
        self.mat_Kp.mult(y, z)  # z = K_p y
        z.axpy(1.0, x)          # z = z + x
        self.ksp_Mp.solve(z, y) # y = M_p^{-1} z
        self.ksp_Rp.solve(x, z) # z = R_p^{-1} x
        y.axpy(1.0, z)          # y = y + z
        # FIXME: How is with the sign bussines?
        y.scale(-1.0)           # y = -y



class PCDRPC_BRM2(BasePCDRPC):
    """This class implements an extension of :py:class:`PCDPC_BRM2`.
    Here we add a reaction term into the preconditioner, so that is becomes
    PCDR (pressure-convection-diffusion-reaction) preconditioner. This
    particular variant is suitable for time-dependent problems, where the
    reaction term arises from the time derivative in the balance of momentum.
    """

    @timed("FENaPack: PCDRPC_BRM2 apply")
    def apply(self, pc, x, y):
        # FIXME: Fix the docstring
        """This method implements the action of the inverse of the approximate
        Schur complement :math:`-\hat{S}^{-1}`, that is

        .. math::

            y = - R_p^{-1} x - (I + A_p^{-1} K_p) M_p^{-1} x.

        where :math:`K_p` is used to denote pressure convection matrix, while
        :math:`R_p` originates in the discretized time derivative. Roughly
        speaking, :math:`R_p^{-1} x` corresponds to additional Laplace solve
        in which the minus Laplacian operator is approximated in a clever way
        based on the *velocity mass matrix* and *pressure gradient operator*.

        Based on experimental evidence, we can say that this particular
        variant performs better than :py:class:`PCDPC_BRM2` (with :math:`K_p`
        enriched by the pressure mass matrix from the discrete time
        derivative) applied to time-dependent problems.
        """
        # Fetch work vector
        z0, z1 = self.get_work_vecs(x, 2)

        # Apply PCD
        self.ksp_Mp.solve(x, y)   # y = M_p^{-1} x
        y.copy(result=z0)         # z0 = y
        self.mat_Kp.mult(z0, z1)  # z1 = K_p z0
        self.bcs_applier(z1)      # apply bcs to z1
        self.ksp_Ap.solve(z1, z0) # z0 = A_p^{-1} z1
        y.axpy(1.0, z0)           # y = y + z0
        self.ksp_Rp.solve(x, z0)  # z0 = R_p^{-1} x
        y.axpy(1.0, z0)           # y = y + z0
        # FIXME: How is with the sign bussines?
        y.scale(-1.0)             # y = -y



from dolfin import MPI, has_lu_solver_method
from petsc4py import PETSc

import functools


def get_default_factor_solver_type(comm):
    """Return first available factor solver type name.
    This is implemened using DOLFIN now."""

    methods_parallel = ("mumps", "superlu_dist", "pastix")
    methods_sequential = ("mumps", "umfpack", "superlu",
                          "superlu_dist", "pastix")

    if isinstance(comm, PETSc.Comm):
        comm = comm.tompi4py()

    if MPI.size(comm) > 1:
        methods = methods_parallel
    else:
        methods = methods_sequential

    for method in methods:
        if has_lu_solver_method(method):
            return method

    raise RuntimeError("Did not find any suitable direct sparse solver in PETSc")


def pc_set_factor_solver_type(pc, type):
    if PETSc.Sys.getVersion()[0:2] <= (3, 8) and PETSc.Sys.getVersionInfo()['release']:
        pc.setFactorSolverPackage(type)
    else:
        pc.setFactorSolverType(type)


def allow_only_one_call(func):
    """Decorator allowing provided instancemethod to
    be called only once. Additional calls raise error."""

    sentinel_variable_name = "___called_{}_".format(func.__name__)

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        # Check if already called and raise
        if getattr(self, sentinel_variable_name, False):
            try:
                name = ".".join((__package__, self.__module__,
                                 self.__class__.__name__, func.__name__))
            except Exception:
                name = func.__name__
            raise RuntimeError("Multiple calls to {} not allowed".format(name))

        # Mark the function as called
        setattr(self, sentinel_variable_name, True)

        # Call the function
        return func(self, *args, **kwargs)

    return wrapper



# -*- coding: utf-8 -*-

# Copyright (C) 2016-2017 Patrick Farrell, Jan Blechta
#
# This file comes originally from defcon.
# This file is part of FENaPack.
#
# FENaPack is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FENaPack is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FENaPack.  If not, see <http://www.gnu.org/licenses/>.

# Let PETSc print memory usage at the end
import petsc4py
petsc4py.init(("python", "-malloc_info"))

from defcon import *
from dolfin import *
from matplotlib import pyplot

from fenapack import PCDKSP
from fenapack import PCDAssembler
from fenapack import StabilizationParameterSD

import sys
import argparse


class NavierStokesProblem(BifurcationProblem):
    def __init__(self):
        # FIXME: Hack for defcon gui
        argv = [] if __name__ == "prob" else sys.argv[1:]

        # Parse input arguments
        parser = argparse.ArgumentParser(description=__doc__, formatter_class=
                                         argparse.ArgumentDefaultsHelpFormatter)
        parser.add_argument("--pcd", type=str, dest="pcd", default="BRM2",
                            choices=["none", "BRM1", "BRM2"], help="PCD variant")
        parser.add_argument("--nls", type=str, dest="nls", default="newton",
                            choices=["picard", "newton"], help="nonlinear solver")
        parser.add_argument("--ls", type=str, dest="pcdls", default="direct",
                            choices=["direct", "iterative"], help="PCD linear solvers")
        parser.add_argument("--ksp_monitor", action="store_true")
        self.args = parser.parse_args(argv)

    def mesh(self, comm):
        mesh = Mesh(comm, "mesh/mesh.xml.gz")
        return mesh

    def function_space(self, mesh):
        Ve = VectorElement("CG", triangle, 2)
        Qe = FiniteElement("CG", triangle, 1)
        Ze = MixedElement([Ve, Qe])
        Z  = FunctionSpace(mesh, Ze)
        return Z

    def parameters(self):
        Re = Constant(0)
        return [(Re, "Re", r"$\mathrm{Re}$")]

    def residual(self, z, params, w):
        (u, p) = split(z)
        (v, q) = split(w)

        Re = params[0]
        mesh = z.function_space().mesh()

        # Variational form
        F = (
              1.0/Re * inner(grad(u), grad(v))*dx
            + inner(grad(u)*u, v)*dx
            - p*div(v)*dx
            - q*div(u)*dx
            )

        # Trial functions
        u_, p_ = TrialFunctions(z.function_space())

        # Jacobian
        if self.args.nls == "picard":
            J = (
                  1.0/Re * inner(grad(u_), grad(v))*dx
                + inner(grad(u_)*u, v)*dx
                - p_*div(v)*dx
                - q*div(u_)*dx
            )
        elif self.args.nls == "newton":
            J = derivative(F, z)

        # Store for later use
        self._J = J

        # If not using PCD we are done
        if self.args.pcd == "none":
            return F

        if self.args.pcdls == "iterative":
            # Add stabilization for AMG 00-block
            iRe = Expression("1./Re", Re=Re, degree=0, mpi_comm=mesh.mpi_comm())
            delta = StabilizationParameterSD(z.sub(0), iRe)
            J_pc = J + delta*inner(grad(u_)*u, grad(v)*u)*dx
            J_pc = None
        else:
            J_pc = None

        # Fetch problem BCs and facet markers
        bcs = self.boundary_conditions(z.function_space(), self.parameters())
        colours = self._colours

        # PCD operators
        mp = Re*p_*q*dx
        kp = Re*dot(grad(p_), u)*q*dx
        ap = inner(grad(p_), grad(q))*dx
        if self.args.pcd == "BRM2":
            n = FacetNormal(mesh)
            ds = Measure("ds", subdomain_data=colours)
            kp -= Re*dot(u, n)*p_*q*ds(1)
            #kp -= Re*dot(u, n)*p_*q*ds(0)  # TODO: Is this beneficial?

        # Artificial BC for PCD preconditioner
        if self.args.pcd == "BRM1":
            bc_pcd = DirichletBC(z.function_space().sub(1), 0.0, colours, 1)
        elif self.args.pcd == "BRM2":
            bc_pcd = DirichletBC(z.function_space().sub(1), 0.0, colours, 2)

        # Store what needed for later
        self._pcd_assembler = PCDAssembler(J, F, bcs, J_pc,
                                           ap=ap, kp=kp, mp=mp,
                                           bcs_pcd=bc_pcd)

        return F

    def boundary_conditions(self, Z, params):
        comm = Z.mesh().mpi_comm()

        # Facet markers
        colours = MeshFunction("size_t", Z.mesh(), Z.mesh().topology().dim()-1)
        walls = CompiledSubDomain("on_boundary", mpi_comm=comm)
        inflow = CompiledSubDomain("on_boundary && near(x[0], 0.0)", mpi_comm=comm)
        outflow = CompiledSubDomain("on_boundary && near(x[0], 150.0)", mpi_comm=comm)
        colours.set_all(999)      # interior facets
        walls.mark(colours, 0)    # no-slip facets
        inflow.mark(colours, 1)   # inlet facets
        outflow.mark(colours, 2)  # outlet facets

        # BCs
        poiseuille = Expression(("-(x[1] + 1) * (x[1] - 1)", "0.0"), degree=2, mpi_comm=comm)
        bc_inflow = DirichletBC(Z.sub(0), poiseuille, colours, 1)
        bc_wall = DirichletBC(Z.sub(0), (0, 0), colours, 0)

        bcs = [bc_inflow, bc_wall]

        # Store markers for later use
        self._colours = colours

        return bcs

    def functionals(self):
        def sqL2(z, params):
            (u, p) = split(z)
            # FIXME: Why is here |z|^2
            j = assemble(inner(z, z)*dx)
            return j

        return [(sqL2, "sqL2", r"$\|u\|^2$")]

    def number_initial_guesses(self, params):
        return 1

    def initial_guess(self, Z, params, n):
        return Function(Z)

    def number_solutions(self, params):
        return float("inf")

        #Re = params[0]
        #if   Re < 18:  return 1
        #elif Re < 41:  return 3
        #elif Re < 75:  return 5
        #elif Re < 100: return 8
        #else:          return float("inf")

    def solver_parameters(self, params, klass, averaging=False):
        opts = {
            "snes_monitor": None,
            "snes_converged_reason": None,
            "snes_max_it": 8,
            "snes_atol": 1.0e-9,
            "snes_rtol": 0.0,
        }

        if averaging:
            opts.update({
                "snes_max_it": 32,
                "snes_linesearch_damping": 0.1,
            })

        if self.args.pcd == "none":
            # Completely direct solver, no PCD
            opts.update({
                "ksp_type": "preonly",
                "pc_type": "lu",
                "pc_factor_mat_solver_package": "mumps",
            })
        else:
            # GMRES with PCD
            opts.update({
                "ksp_converged_reason": None,
                "ksp_max_it": 128,
                "ksp_rtol": 1.0e-5,
                "ksp_atol": 0.0,
                "snes_max_linear_solve_fail": 4,
                "ksp_gmres_restart": 128,
                "fieldsplit_p_pc_python_type": "fenapack.PCDPC_"+self.args.pcd,
            })

            if self.args.ksp_monitor:
                opts.update({"ksp_monitor": None})

            if self.args.pcdls == "iterative":
                # Iterative inner PCD solves
                opts.update({
                    "snes_max_it": 16,  # FIXME: Can we improve this?
                    "fieldsplit_u_ksp_type": "richardson",
                    "fieldsplit_u_ksp_max_it": 1,
                    "fieldsplit_u_pc_type": "hypre",
                    "fieldsplit_u_pc_hypre_type": "boomeramg",
                    "fieldsplit_p_PCD_Ap_ksp_type": "richardson",
                    "fieldsplit_p_PCD_Ap_ksp_max_it": 2,
                    "fieldsplit_p_PCD_Ap_pc_type": "hypre",
                    "fieldsplit_p_PCD_Ap_pc_hypre_type": "boomeramg",
                    "fieldsplit_p_PCD_Mp_ksp_type": "chebyshev",
                    "fieldsplit_p_PCD_Mp_ksp_max_it": 5,
                    # FIXME: Only valid in 2D (see SEW book)?
                    "fieldsplit_p_PCD_Mp_ksp_chebyshev_eigenvalues": "0.5, 2.0",
                    "fieldsplit_p_PCD_Mp_pc_type": "jacobi",
                })

        return opts

    def jacobian(self, F, state, params, test, trial):
        return self._J

    def solver(self, problem, params, solver_params, prefix="", **kwargs):
        # Create nonlinear solver
        solver = SNUFLSolver(problem, prefix=prefix,
                             solver_parameters=solver_params,
                             **kwargs)

        # This is enough for completely direct solve without PCD
        if self.args.pcd == "none":
            return solver

        # Create GMRES KSP with PCD
        ksp = PCDKSP(comm=problem.u.function_space().mesh().mpi_comm())

        # Switch ksp and reuse operators
        oldksp = solver.snes.ksp
        ksp.setOperators(*oldksp.getOperators())
        ksp.setOptionsPrefix(oldksp.getOptionsPrefix())
        solver.snes.ksp = ksp
        solver.snes.setFromOptions()

        # Initilize PCD (only possible with assembled operators)
        # NOTE: It is important that we call jacobian() with x, not a dummy
        #       vector, so that x is updated correctly for continuation
        x = as_backend_type(problem.u.vector()).vec()
        # FIXME: Make sure this works when J_pc is not used
        solver.jacobian(solver.snes, x, *solver.snes.ksp.getOperators())
        ksp.init_pcd(self._pcd_assembler)

        return solver


if __name__ == "__main__":
    dc = DeflatedContinuation(problem=NavierStokesProblem(), teamsize=1, verbose=True)
    #dc.run(values={"Re": linspace(10.0, 100.0, 181)})
    dc.run(values={"Re": arange(18.0, 20.5, 0.5)})

    # FIXME: This is not what we want possibly; an average over WORLD,
    #        thus biased by master thread
    list_timings(TimingClear.keep, [TimingType.wall, TimingType.user])

    dc.bifurcation_diagram("sqL2")
    pyplot.title(r"Bifurcation diagram for sudden expansion in a channel")
    pyplot.savefig("bifurcation.pdf")



# -*- coding: utf-8 -*-

# Copyright (C) 2014-2017 Jan Blechta and Martin Řehoř
#
# This file is part of FENaPack.
#
# FENaPack is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FENaPack is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FENaPack.  If not, see <http://www.gnu.org/licenses/>.

"""This is FENaPack, FEniCS Navier-Stokes preconditioning package."""

__author__ = "Jan Blechta, Martin Řehoř"
__version__ = "2019.1.0"
__license__ = "GNU LGPL v3"

# Do not use petsc4py python error handler (hides error messages
# to PETSc C calls), workaround to DOLFIN issue #801
from dolfin import SubSystemsManager
SubSystemsManager.init_petsc()  # init PETSc by DOLFIN before petsc4py import
from petsc4py import PETSc
PETSc.Sys.pushErrorHandler("traceback")
del SubSystemsManager, PETSc

# Import public API
from fenapack.field_split import PCDKSP, PCDKrylovSolver
from fenapack.assembling import PCDAssembler, PCDForm
from fenapack.nonlinear_solvers import PCDNewtonSolver, PCDNonlinearProblem
from fenapack.preconditioners import PCDPC_BRM1, PCDPC_BRM2
from fenapack.preconditioners import PCDRPC_BRM1, PCDRPC_BRM2
from fenapack.stabilization import StabilizationParameterSD


# Monkey patch dolfin.NewtonSolver in 2018.1.0
from dolfin import NewtonSolver, compile_cpp_code
_cpp = """
#include <pybind11/pybind11.h>
#include <dolfin/nls/NewtonSolver.h>

PYBIND11_MODULE(SIGNATURE, m)
{
  pybind11::class_<dolfin::NewtonSolver, std::shared_ptr<dolfin::NewtonSolver>>
  (m, "NewtonSolverExt", pybind11::module_local())
  .def("krylov_iterations", &dolfin::NewtonSolver::krylov_iterations);
}
"""
NewtonSolver.krylov_iterations = compile_cpp_code(_cpp).NewtonSolverExt.krylov_iterations
del _cpp, NewtonSolver, compile_cpp_code



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from  setuptools import setup
import re, os, glob

version = re.findall('__version__ = "(.*)"',
                     open('fenapack/__init__.py', 'r').read())[0]

CLASSIFIERS = """
Development Status :: 2 - Pre-Alpha
Environment :: Console
Intended Audience :: Science/Research
License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)
Programming Language :: Python :: 3
Programming Language :: C++
Topic :: Scientific/Engineering :: Mathematics
"""
classifiers = CLASSIFIERS.split('\n')[1:-1]

demofiles = (
      glob.glob(os.path.join("demo", "**", "*.py"), recursive=True)
    + glob.glob(os.path.join("demo", "**", "*.rst"), recursive=True)
    + glob.glob(os.path.join("demo", "data", "*.xml"))
)
data_files=[(os.path.join("share", "fenapack", os.path.dirname(f)), [f])
            for f in demofiles]

with open("README.rst", "r") as f:
    long_description = f.read()

setup(name="fenapack",
      version=version,
      author="Jan Blechta",
      author_email="blechta@karlin.mff.cuni.cz",
      url="https://github.com/blechta/fenapack",
      description="FENaPack: FEniCS Navier-Stokes preconditioning package",
      long_description=long_description,
      long_description_content_type="text/x-rst",
      classifiers=classifiers,
      packages=["fenapack"],
      package_dir={"fenapack": "fenapack"},
      package_data={"fenapack": ["*.h"]},
      data_files=data_files)



# Copyright (C) 2014-2017 Jan Blechta and Martin Rehor
#
# This file is part of FENaPack.
#
# FENaPack is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FENaPack is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with FENaPack.  If not, see <http://www.gnu.org/licenses/>.

"""This module provides subclasses of DOLFIN and petsc4py
Krylov solvers implementing PCD fieldsplit preconditioned
GMRES"""

from __future__ import print_function

from dolfin import Timer, PETScKrylovSolver
from petsc4py import PETSc
from mpi4py import MPI

from fenapack._field_split_utils import dofmap_dofs_is
from fenapack.field_split_backend import PCDInterface
from fenapack.preconditioners import PCDPC_BRM1
from fenapack.utils import get_default_factor_solver_type
from fenapack.utils import pc_set_factor_solver_type
from fenapack.utils import allow_only_one_call


class PCDKSP(PETSc.KSP):
    """GMRES with right fieldsplit preconditioning using upper
    Schur factorization and PCD Schur complement approximation.

    This is a subclass of ``petsc4py.KSP``."""

    # NOTE: We are not able to overload PETSc.KSP methods in this
    # class. That would only be possible with KSPPYTHON type but
    # we do not need it.

    def __init__(self, comm=None):
        """Initialize PCDKSP on given MPI comm"""

        super(PCDKSP, self).__init__()

        self.create(comm)
        self.setType(PETSc.KSP.Type.GMRES)
        self.setPCSide(PETSc.PC.Side.RIGHT)
        self.pc.setType(PETSc.PC.Type.FIELDSPLIT)
        self.pc.setFieldSplitType(PETSc.PC.CompositeType.SCHUR)
        self.pc.setFieldSplitSchurFactType(PETSc.PC.SchurFactType.UPPER)
        self.pc.setFieldSplitSchurPreType(PETSc.PC.SchurPreType.USER)


    @allow_only_one_call
    def init_pcd(self, pcd_assembler, pcd_pc_class=None):
        """Initialize from ``PCDAssembler`` instance. Needs to be called
        after ``setOperators`` and ``setUp``. That's why two-phase
        initialization is needed: first ``__init__``, then ``init_pcd``

        Note that this function automatically calls setFromOptions to
        all subKSP objects.
        """

        # Get subfield index sets
        V = pcd_assembler.function_space()
        is0 = dofmap_dofs_is(V.sub(0).dofmap())
        is1 = dofmap_dofs_is(V.sub(1).dofmap())

        comm = self.comm.tompi4py()
        assert comm.Compare(comm, V.mesh().mpi_comm()) in \
            [MPI.IDENT, MPI.CONGRUENT], "Non-matching MPI comm"

        # Set subfields index sets
        # NOTE: Doing only so late here so that user has a chance to
        # set options prefix, see PETSc issue #160
        self.pc.setFieldSplitIS(["u", is0], ["p", is1])

        # From now on forbid setting options prefix (at least from Python)
        self.setOptionsPrefix = self._forbid_setOptionsPrefix

        # Setup fieldsplit preconditioner
        pc_prefix = self.pc.getOptionsPrefix() or ""
        with Timer("FENaPack: PCDKSP PC {} setup".format(pc_prefix)):
            self.pc.setUp()

        # Extract fieldsplit subKSPs (only once self.pc is set up)
        ksp0, ksp1 = self.pc.getFieldSplitSubKSP()

        # Set some sensible defaults
        ksp0.setType(PETSc.KSP.Type.PREONLY)
        ksp0.pc.setType(PETSc.PC.Type.LU)
        pc_set_factor_solver_type(ksp0.pc, get_default_factor_solver_type(comm))
        ksp1.setType(PETSc.KSP.Type.PREONLY)
        ksp1.pc.setType(PETSc.PC.Type.PYTHON)

        # Setup 0,0-block pc so that we have accurate timing
        ksp0_prefix = ksp0.getOptionsPrefix()
        with Timer("FENaPack: {} setup".format(ksp0_prefix)):
            ksp0.setFromOptions()  # Override defaults above by user's options
            ksp0.pc.setUp()

        # Initialize PCD PC context
        pcd_pc_prefix = ksp1.pc.getOptionsPrefix()
        pcd_pc_opt = PETSc.Options(pcd_pc_prefix).getString("pc_python_type","")
        # Use PCDPC class given by option
        if pcd_pc_opt != "":
            ksp1.setFromOptions()  # Override defaults above by user's options
            pcd_pc = ksp1.pc.getPythonContext()
        # Use PCDPC class specified as argument
        elif pcd_pc_class is not None:
            pcd_pc = pcd_pc_class()
            ksp1.pc.setPythonContext(pcd_pc)
            ksp1.setFromOptions()  # Override defaults above by user's options
        # Use default PCDPC class
        else:
            pcd_pc = PCDPC_BRM1()
            ksp1.pc.setPythonContext(pcd_pc)
            ksp1.setFromOptions()  # Override defaults above by user's options

        # Get backend implementation of PCDAssembler
        # FIXME: Make me parameter
        #deep_submats = False
        deep_submats = True
        A = self.getOperators()[0]
        pcd_interface = PCDInterface(pcd_assembler, A, is0, is1,
                                     deep_submats=deep_submats)

        # Provide assembling routines to PCD
        try:
            pcd_pc.init_pcd(pcd_interface)
        except Exception:
            print("Initialization of PCD PC from PCDAssembler failed!")
            print("Maybe wrong PCD PC class or PCDAssembler instance.")
            raise

        # Setup PCD PC so that we have accurate timing
        with Timer("FENaPack: {} setup".format(pcd_pc_prefix)):
            ksp1.pc.setUp()


    def _forbid_setOptionsPrefix(self, prefix):
        raise RuntimeError("Options prefix cannot be set now. "
                           "Set it before init_pcd.")



class PCDKrylovSolver(PETScKrylovSolver):
    """GMRES with right fieldsplit preconditioning using upper
    Schur factorization and PCD Schur complement approximation.

    This is a subclass of ``dolfin.PETScKrylovSolver``."""

    def __init__(self, comm=None):
        """Initialize Krylov solver on given MPI comm"""
        self._ksp = PCDKSP(comm=comm)
        super(PCDKrylovSolver, self).__init__(self._ksp)


    def init_pcd(self, pcd_assembler, pcd_pc_class=None):
        """Initialize from ``PCDAssembler`` instance. Needs to be called
        after ``setOperators`` and ``setUp``. That's why two-phase
        initialization is needed: first ``__init__``, then ``init_pcd``

        Note that this function automatically calls setFromOptions to
        all subKSP objects.
        """
        self._ksp.init_pcd(pcd_assembler, pcd_pc_class=pcd_pc_class)


    def ksp(self):
        return self._ksp


    ksp.__doc__ = PETScKrylovSolver.ksp.__doc__


    def set_options_prefix(self, prefix):
        self._ksp.setOptionsPrefix(prefix)


    set_options_prefix.__doc__ = PETScKrylovSolver.set_options_prefix.__doc__
